(define (create-account initial-balance)
  (let ((balance initial-balance))
    (lambda (method . args)
      (cond ((equal? method 'deposit) (set! balance (+ balance (car args))))
	    ((equal? method 'withdraw) (set! balance (- balance (car args))))
	    ((equal? method 'balance) balance)))))

(define (deposit! account amount)
  (account 'deposit amount))
(define (withdraw! account amount)
  (account 'withdraw amount))
(define (balance account)
  (account 'balance))


(define (create-bank accounts)
  (let loop ((i accounts))
    (if (zero? i) '()
      (cons (cons i (create-account (* i 1000)))
	    (loop (- i 1))))))

(define (bank-get-account bank id)
  (cond ((null? bank) '())
	((equal? (caar bank) id) (cdar bank))
	(else (bank-get-account (cdr bank) id))))


(define (transfer! bank id-1 id-2 amount)
  (let ((account-1 (bank-get-account bank id-1))
	(account-2 (bank-get-account bank id-2)))
    (if (> (balance account-1) amount)
      (begin
;	(display "transfer")
;	(display id-1)
;	(display "->")
;	(display id-2)
;	(display ":")
;	(display amount)
;	(newline)
	(withdraw! account-1 amount)
	(deposit! account-2 amount))
      0
      )))


(define (create-transaction-list accounts length)
  (let loop ((i length))
    (if (zero? i) '()
      (let ((id-1 (+ (modulo (* i 31) accounts) 1))
	    (id-2 (+ (modulo (* i 56) accounts) 1)))
	(cons (list id-1 id-2 (+ id-1 id-2))
	      (loop (- i 1)))))))


(define (caddar x) (car (cddar x)))
(define (test accounts transactions)
  (define bank (create-bank accounts))
  (define (loop transactions)
   (if (null? transactions) 'done
     (begin
       (transfer! bank (caar transactions) (cadar transactions) (caddar transactions))
       (loop (cdr transactions)))))
  (loop (create-transaction-list accounts transactions))
;  (map (lambda (x)
;	 (transfer! bank (car x) (cadr x) (caddr x)))
;    (create-transaction-list accounts transactions))

  (map (lambda (x)
	 (map display (list (car x) ":" (balance (cdr x))))
	 (newline))
      bank)
  )

;(test 120 5000)
(test 200 1000)
