;;
;;
;; Basic functions
;;
;;


;;
;; List functions
;;

  (define (x-map fun lst) 
    (if (null? lst) 
      lst 
      (cons (fun (car lst)) (map fun (cdr lst)))))

;(define map map-exp)

;(define (map func . args)
;  (define (head-n-tails args heads tails)
;    (if (null? args) (cons heads tails)
;      (head-n-tails (cdr args)
;		    (cons (caar args) heads)
;		    (cons (cdar args) tails))))
;  (define h-n-t (head-n-tails args '() '()))
;  (if (null? (cdar args)) (list (apply func (car h-n-t)))
;    (cons (apply func (car h-n-t)) (apply map func (cdr h-n-t)))))
;

(define (caar x) (car (car x)))
(define (cadr x) (car (cdr x)))
(define (cdar x) (cdr (car x)))
(define (cddr x) (cdr (cdr x)))
(define (caaar x) (car (car (car x))))
(define (cadar x) (car (cdr (car x))))
(define (caadr x) (car (car (cdr x))))
(define (cddar x) (cdr (cdr (car x))))
(define (caddr x) (car (cdr (cdr x))))
(define (cdadr x) (cdr (car (cdr x))))
(define (cdddr x) (cdr (cdr (cdr x))))

(define first car)
(define second cadr)
(define third caddr)

;(define (assoc item list)
;    (cond ((null? list) #f)
;	  ((equal? item (caar list)) (car list))
;	  (else (assoc item (cdr list)))))

(define assoc #f)
(define assv #f)
(define assq #f)
(letrec ((generic-assoc
	   (lambda (item list test-fn)
	     (cond ((null? list) #f)
		   ((test-fn item (caar list)) (car list))
		   (else (generic-assoc item (cdr list) test-fn))))))
  (set! assoc (lambda (item list) (generic-assoc item list equal?)))
  (set! assv (lambda (item list) (generic-assoc item list eqv?)))
  (set! assq (lambda (item list) (generic-assoc item list eq?))))

(define memq #f)
(define memv #f)
(letrec ((generic-member
	   (lambda (item list test-fn)
	     (cond ((null? list) #f)
		   ((test-fn item (car list)) list)
		   (else (generic-member item (cdr list) test-fn))))))
  (set! memv (lambda (item list) (generic-member item list eqv?)))
  (set! memq (lambda (item list) (generic-member item list eq?))))



(define (list-tail ls k)
    (if (= k 0) ls (list-tail (cdr ls) (- k 1))))

(define (list-ref ls k)
    (car (list-tail ls k)))

(define (<= x y) (not (> x y)))
(define (>= x y) (not (< x y)))

(define (displayn arg)
	(display arg)
	(newline))

(define define-macro
  (macro (lambda (first . rest)
	   (if (pair? first)
	     `(define ,(car first) (macro (lambda ,(cdr first) ,@rest)))
	     `(define ,first (macro ,@rest))))))

(define (zero? x) (= x 0))
(define (positive? x) (> x 0))
(define (negative? x) (< x 0))
(define (even? x) (= (modulo x 2) 0))
(define (odd? x) (= (modulo x 2) 1))
(define (abs x)
  (if (< x 0) (- x) x))

(define pi 3.1415926535897932384)

;(define (append! x y)
;  (if (null? x) y
;    (let loop ((z x))
;      (if (null? (cdr z)) 
;	(begin
;	  (set-cdr! z y)
;	  x)
;	(loop (cdr z))))))

(define void (lambda x (if #f #f)))

;;
;; shell-related functions
;;

(define (ls . args)
  (apply ext:shell* 'LIST "ls" '() args))

(define (regexp:grep pattern list)
  (ext:shell-unsync* 'LIST "grep" list "-e" pattern))

(define (cut field delim list)
  (ext:shell-unsync* 'LIST "cut" list "-f" field "-d" delim))

(define (string-sort list)
  (ext:shell-unsync* 'LIST "sort" list))

(define (file->list file . files)
  (apply ext:shell* 'LIST "cat" "" file files))

(define (file->string file . files)
  (apply ext:shell* 'STRING "cat" "" file files))

(define (pwd)
  (ext:shell* 'STRING "pwd"))

;;
;; lazy-evaluation
;;

(define-macro (delay expr)
	      `(let ((initialized #f)
		     (value #f))
		 (lambda ()
		   (if initialized value
		     (begin
		       (set! initialized #t)
		       (set! value ,expr)
		       value)))))

(define (force promise)
  (promise))

;; R5RS compatibility

; `for-each' and `map' are equivalent in schemi
(define for-each map)
(define symbol->string object->string)
(define number->string object->string)
(define (string=? x y)
  (if (and (string? x) (string? y))
    (equal? x y)
    (error "Function string=? expects STRINGs as its arguments")))

(define complex? number?)
(define real? number?)
(define (integer? x) (= (truncate x) x))

(define-macro (case key . options)
   (let* ((key-value (gensym))
	  (make-clause (lambda (x)
			(if (list? (car x))
			  `((or ,@(map (lambda (y) `(eqv? ,key-value ',y))
				      	(car x)))
			    ,@(cdr x))
			  x))))
     `(let ((,key-value ,key))
	(cond ,@(map make-clause options)))))

(define-macro (do init test . commands)
  (let ((loop-name (gensym)))
    `(let ,loop-name 
       ,(map (lambda (i) (list (car i) (cadr i))) init)
       (if ,(car test) (begin ,@(cdr test))
	 (begin ,@commands
		(,loop-name ,@(map (lambda (i) 
				     (if (null? (cddr i)) (car i) (caddr i)))
				   init)))))))
