(define (iterator state)
  (lambda (x)
    (cond ((equal? x 'next) (let ((a (car state)))
			      (set! state (cdr state))
			      a))
	  ((equal? x 'has-next) (not (null? state)))
	  (else "unknown signal"))))

(define (iter-next! it) (it 'next))
(define (iter-has-next? it) (it 'has-next))

(define (aggregator)
  (let ((state '()))
    (lambda (x . value)
      (cond ((equal? x 'add) (set! state (cons (car value) state)))
	    ((equal? x 'state) (reverse state))
	    (else "unknown signal")))))

(define (aggr-add! ag val) (ag 'add val))
(define (aggr-state ag) (ag 'state))

(define (sum-up list-1 list-2)
  (let ((it-1 (iterator list-1))
	(it-2 (iterator list-2))
	(result (aggregator)))
    (let loop ()
      (if (and (iter-has-next? it-1)
	       (iter-has-next? it-2))
	(begin
	  (aggr-add! result (+ (iter-next! it-1) (iter-next! it-2)))
	  (loop))
	(aggr-state result)))))



(define (seq from to)
  (if (= from to) (list to) (cons from (seq (+ from 1) to))))

(define (matrix-create n value)
  (define (iter m v)
    (if (= m 0) () (cons v (iter (- m 1) v))))
  (iter n (iter n value)))

;(sum-up (seq 1 100000) (seq 1 100000))

;(define (add m n)
;    (map (lambda (r1 r2)
;	            (map (lambda (c1 c2) (+ c1 c2)) r1 r2))
;	          m n))

(define (b-loop cnt)
  (let loop ((i cnt))
    (if (> i 0) (loop (- i 1)))))

(define (gen-op f m n)
  (let ((it-1 (iterator m))
	(it-2 (iterator n))
	(result (aggregator)))
    (let loop ()
      (if (and (iter-has-next? it-1)
	       (iter-has-next? it-2))
	(begin
	  (aggr-add! result
		     (f (iter-next! it-1) (iter-next! it-2)))

	  (loop))
	(aggr-state result)))))

(define (add m n)
  (gen-op (lambda (x y)
	    (gen-op * x y))
	  m n))


(define SIZE 200)
(define A (matrix-create SIZE 1))
(define B (matrix-create SIZE 4))
(add A B)



